from functools import lru_cache

from cms.models import CMSPlugin
from cms.plugin_pool import plugin_pool


@lru_cache()
def get_local_fields(model):
    opts = model._meta.concrete_model._meta
    fields = opts.local_fields
    return [
        field.name
        for field in fields
        if not field.is_relation and not field.primary_key
    ]


@lru_cache()
def get_related_fields(model):
    opts = model._meta.concrete_model._meta
    fields = opts.local_fields + list(opts.many_to_many)
    return [field.name for field in fields if field.is_relation]


@lru_cache()
def get_plugin_class(plugin_type):
    return plugin_pool.get_plugin(plugin_type)


@lru_cache()
def get_plugin_fields(plugin_type):
    klass = get_plugin_class(plugin_type)
    if klass.model is CMSPlugin:
        return []
    opts = klass.model._meta.concrete_model._meta
    fields = opts.local_fields + opts.local_many_to_many
    return [field.name for field in fields]


@lru_cache()
def get_plugin_model(plugin_type):
    return get_plugin_class(plugin_type).model



import functools
import json

from django.core.serializers.json import DjangoJSONEncoder

from . import helpers


dump_json = functools.partial(json.dumps, cls=DjangoJSONEncoder)


def export_plugin(plugin):
    data = get_plugin_export_data(plugin)
    return dump_json(data)


def export_placeholder(placeholder, language):
    data = get_placeholder_export_data(placeholder, language)
    return dump_json(data)


def export_page(cms_pagecontent, language):
    data = get_page_export_data(cms_pagecontent, language)
    return dump_json(data)


def get_plugin_export_data(plugin):
    get_data = helpers.get_plugin_data
    descendants = plugin.get_descendants()
    plugin_data = [get_data(plugin=plugin)]
    plugin_data[0]["parent_id"] = None
    plugin_data.extend(
        get_data(plugin) for plugin in helpers.get_bound_plugins(descendants)
    )
    return plugin_data


def get_placeholder_export_data(placeholder, language):
    get_data = helpers.get_plugin_data
    plugins = placeholder.get_plugins(language)
    # The following results in two queries;
    # First all the root plugins are fetched, then all child plugins.
    # This is needed to account for plugin path corruptions.

    return [get_data(plugin) for plugin in helpers.get_bound_plugins(list(plugins))]


def get_page_export_data(cms_pagecontent, language):
    data = []
    placeholders = cms_pagecontent.rescan_placeholders().values()

    for placeholder in list(placeholders):
        plugins = get_placeholder_export_data(placeholder, language)
        data.append({"placeholder": placeholder.slot, "plugins": plugins})
    return data



import json

from django.core.exceptions import PermissionDenied
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import render
from django.urls import re_path, reverse
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase, PluginMenuItem
from cms.plugin_pool import plugin_pool
from cms.utils import get_language_from_request
from cms.utils.urlutils import admin_reverse

from .forms import ExportImportForm, PluginExportForm, PluginImportForm


class PluginImporter(CMSPluginBase):
    system = True
    render_plugin = False

    def get_plugin_urls(self):
        urlpatterns = [
            re_path(
                r"^export-plugins/$",
                self.export_plugins_view,
                name="cms_export_plugins",
            ),
            re_path(
                r"^import-plugins/$",
                self.import_plugins_view,
                name="cms_import_plugins",
            ),
        ]
        return urlpatterns

    def get_extra_placeholder_menu_items(self, request, placeholder):  # noqa
        # django-cms 3.4 compatibility
        return self.get_extra_placeholder_menu_items(request, placeholder)

    def get_extra_global_plugin_menu_items(self, request, plugin):
        # django-cms 3.4 compatibility
        return self.get_extra_plugin_menu_items(request, plugin)

    @classmethod
    def get_extra_plugin_menu_items(cls, request, plugin):
        if plugin.plugin_type == cls.__name__:
            return

        data = urlencode(
            {
                "language": get_language_from_request(request),
                "plugin": plugin.pk,
            }
        )
        return [
            PluginMenuItem(
                _("Export plugins"),
                admin_reverse("cms_export_plugins") + "?" + data,
                data={},
                action="none",
                attributes={
                    "icon": "export",
                },
            ),
            PluginMenuItem(
                _("Import plugins"),
                admin_reverse("cms_import_plugins") + "?" + data,
                data={},
                action="modal",
                attributes={
                    "icon": "import",
                },
            ),
        ]

    @classmethod  # noqa
    def get_extra_placeholder_menu_items(cls, request, placeholder):  # noqa
        data = urlencode(
            {
                "language": get_language_from_request(request),
                "placeholder": placeholder.pk,
            }
        )
        return [
            PluginMenuItem(
                _("Export plugins"),
                admin_reverse("cms_export_plugins") + "?" + data,
                data={},
                action="none",
                attributes={
                    "icon": "export",
                },
            ),
            PluginMenuItem(
                _("Import plugins"),
                admin_reverse("cms_import_plugins") + "?" + data,
                data={},
                action="modal",
                attributes={
                    "icon": "import",
                },
            ),
        ]

    @classmethod
    def import_plugins_view(cls, request):
        if not request.user.is_staff:
            raise PermissionDenied

        new_form = ExportImportForm(request.GET or None)

        if new_form.is_valid():
            initial_data = new_form.cleaned_data
        else:
            initial_data = None

        if request.method == "GET" and not new_form.is_valid():
            return HttpResponseBadRequest(_("Form received unexpected values."))

        import_form = PluginImportForm(
            data=request.POST or None,
            files=request.FILES or None,
            initial=initial_data,
        )

        if not import_form.is_valid():
            opts = cls.model._meta
            context = {
                "form": import_form,
                "has_change_permission": True,
                "opts": opts,
                "root_path": reverse("admin:index"),
                "is_popup": True,
                "app_label": opts.app_label,
                "media": (cls().media + import_form.media),
            }
            return render(request, "djangocms_transfer/import_plugins.html", context)

        plugin = import_form.cleaned_data.get("plugin")
        language = import_form.cleaned_data["language"]

        if plugin:
            root_id = plugin.pk
            placeholder = plugin.placeholder
        else:
            root_id = None
            placeholder = import_form.cleaned_data.get("placeholder")

        if not placeholder:
            # Page placeholders/plugins import
            # TODO: Check permissions
            import_form.run_import()
            return HttpResponse(
                '<div><div class="messagelist"><div class="success"></div></div></div>'
            )

        tree_order = placeholder.get_plugin_tree_order(language, parent_id=root_id)
        # TODO: Check permissions
        import_form.run_import()

        if plugin:
            new_plugins = plugin.reload().get_descendants().exclude(pk__in=tree_order)
            return plugin.get_plugin_class_instance().render_close_frame(
                request, obj=new_plugins[0]
            )

        from cms.toolbar.utils import get_plugin_tree_as_json

        # Placeholder plugins import
        new_plugins = placeholder.get_plugins(language).exclude(pk__in=tree_order)
        data = json.loads(get_plugin_tree_as_json(request, list(new_plugins)))
        data["plugin_order"] = tree_order + ["__COPY__"]
        data["target_placeholder_id"] = placeholder.pk
        context = {"structure_data": json.dumps(data)}
        return render(
            request, "djangocms_transfer/placeholder_close_frame.html", context
        )

    @classmethod
    def export_plugins_view(cls, request):
        if not request.user.is_staff:
            raise PermissionDenied

        form = PluginExportForm(request.GET or None)

        if not form.is_valid():
            return HttpResponseBadRequest(_("Form received unexpected values."))

        # TODO: Check permissions
        filename = form.get_filename()
        response = HttpResponse(form.run_export(), content_type="application/json")
        response["Content-Disposition"] = "attachment; filename={}".format(filename)
        return response


plugin_pool.register_plugin(PluginImporter)



from django.db import transaction

from cms.models import CMSPlugin

from .utils import get_plugin_class


@transaction.atomic
def import_plugins(plugins, placeholder, language, root_plugin_id=None):
    source_map = {}
    new_plugins = []

    if root_plugin_id:
        root_plugin = CMSPlugin.objects.get(pk=root_plugin_id)
        source_map[root_plugin_id] = root_plugin
    else:
        root_plugin = None

    for archived_plugin in plugins:
        # custom handling via "get_plugin_data" can lead to "null"-values
        # instead of plugin-dictionaries. We skip those here.
        if archived_plugin is None:
            continue

        if archived_plugin.parent_id:
            parent = source_map[archived_plugin.parent_id]
        else:
            parent = root_plugin

        if parent and parent.__class__ != CMSPlugin:
            parent = parent.cmsplugin_ptr

        plugin = archived_plugin.restore(
            placeholder=placeholder,
            language=language,
            parent=parent,
        )
        source_map[archived_plugin.pk] = plugin

        new_plugins.append(plugin)

    for new_plugin in new_plugins:
        plugin_class = get_plugin_class(new_plugin.plugin_type)

        if getattr(plugin_class, "_has_do_post_copy", False):
            # getattr is used for django CMS 3.4 compatibility
            # apps on 3.4 wishing to leverage this callback will need
            # to manually set the _has_do_post_copy attribute.
            plugin_class.do_post_copy(new_plugin, source_map)


@transaction.atomic
def import_plugins_to_page(placeholders, page, language):
    page_placeholders = page.rescan_placeholders(language)

    for archived_placeholder in placeholders:
        plugins = archived_placeholder.plugins
        placeholder = page_placeholders.get(archived_placeholder.slot)

        if placeholder and plugins:
            import_plugins(plugins, placeholder, language)



from django.utils.http import urlencode
from django.utils.translation import gettext

from cms.models import PageContent
from cms.toolbar_base import CMSToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils.page_permissions import user_can_change_page
from cms.utils.urlutils import admin_reverse


@toolbar_pool.register
class PluginImporter(CMSToolbar):
    class Media:
        css = {"all": ("djangocms_transfer/css/transfer.css",)}

    def populate(self):
        # always use draft if we have a page
        page = self.request.current_page

        if not page:
            return

        if not user_can_change_page(self.request.user, page):
            return

        page_menu = self.toolbar.get_menu("page")

        if not page_menu or page_menu.disabled:
            return

        obj = self.toolbar.get_object()
        if not obj:
            return
        if not isinstance(obj, PageContent):
            return

        data = urlencode(
            {
                "language": self.current_lang,
                "cms_pagecontent": obj.pk,
            }
        )

        not_edit_mode = not self.toolbar.edit_mode_active

        page_menu.add_break("Page menu importer break")
        page_menu.add_link_item(
            gettext("Export"),
            url=admin_reverse("cms_export_plugins") + "?" + data,
            disabled=not_edit_mode,
        )
        page_menu.add_modal_item(
            gettext("Import"),
            url=admin_reverse("cms_import_plugins") + "?" + data,
            disabled=not_edit_mode,
            on_close=getattr(self.toolbar, "request_path", self.request.path),
        )



import json

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from cms.models import CMSPlugin, PageContent, Placeholder

from .datastructures import ArchivedPlaceholder, ArchivedPlugin
from .exporter import export_page, export_placeholder, export_plugin
from .importer import import_plugins, import_plugins_to_page


def _object_version_data_hook(data, for_page=False):
    if not data:
        return data

    if "plugins" in data:
        return ArchivedPlaceholder(
            slot=data["placeholder"],
            plugins=data["plugins"],
        )

    if "plugin_type" in data:
        return ArchivedPlugin(**data)
    return data


def _get_parsed_data(file_obj, for_page=False):
    raw = file_obj.read().decode("utf-8")
    return json.loads(raw, object_hook=_object_version_data_hook)


class ExportImportForm(forms.Form):
    plugin = forms.ModelChoiceField(
        CMSPlugin.objects.all(),
        required=False,
        widget=forms.HiddenInput(),
    )
    placeholder = forms.ModelChoiceField(
        queryset=Placeholder.objects.all(),
        required=False,
        widget=forms.HiddenInput(),
    )
    cms_pagecontent = forms.ModelChoiceField(
        queryset=PageContent.admin_manager.latest_content(),
        required=False,
        widget=forms.HiddenInput(),
    )
    language = forms.ChoiceField(
        choices=settings.LANGUAGES,
        required=True,
        widget=forms.HiddenInput(),
    )

    def clean(self):
        if self.errors:
            return self.cleaned_data

        plugin = self.cleaned_data.get("plugin")
        placeholder = self.cleaned_data.get("placeholder")
        cms_pagecontent = self.cleaned_data.get("cms_pagecontent")

        if not any([plugin, placeholder, cms_pagecontent]):
            message = _("A plugin, placeholder or page is required.")
            raise forms.ValidationError(message)

        if cms_pagecontent and (plugin or placeholder):
            message = _(
                "Plugins can be imported to pages, plugins or placeholders. Not all three."
            )
            raise forms.ValidationError(message)

        if placeholder and (cms_pagecontent or plugin):
            message = _(
                "Plugins can be imported to pages, plugins or placeholders. Not all three."
            )
            raise forms.ValidationError(message)

        if plugin and (cms_pagecontent or placeholder):
            message = _(
                "Plugins can be imported to pages, plugins or placeholders. Not all three."
            )
            raise forms.ValidationError(message)

        if plugin:
            plugin_model = plugin.get_plugin_class().model
            plugin_is_bound = plugin_model.objects.filter(cmsplugin_ptr=plugin).exists()
        else:
            plugin_is_bound = False

        if plugin and not plugin_is_bound:
            raise ValidationError("Plugin is unbound.")
        return self.cleaned_data


class PluginExportForm(ExportImportForm):
    def get_filename(self):
        data = self.cleaned_data
        language = data["language"]
        cms_pagecontent = data["cms_pagecontent"]
        plugin = data["plugin"]
        placeholder = data["placeholder"]

        if cms_pagecontent:
            return "{}.json".format(cms_pagecontent.page.get_slug(language=language))
        elif placeholder and placeholder.page is not None:
            return "{}_{}.json".format(
                placeholder.page.get_slug(language=language),
                slugify(placeholder.slot),
            )
        elif plugin is not None and plugin.page is not None:
            return "{}_{}.json".format(
                plugin.page.get_slug(language=language),
                slugify(plugin.get_short_description()),
            )
        else:
            return "plugins.json"

    def run_export(self):
        data = self.cleaned_data
        language = data["language"]
        plugin = data["plugin"]
        placeholder = data["placeholder"]

        if plugin:
            return export_plugin(plugin.get_bound_plugin())

        if placeholder:
            return export_placeholder(placeholder, language)
        return export_page(data["cms_pagecontent"], language)


class PluginImportForm(ExportImportForm):
    import_file = forms.FileField(required=True)

    def clean(self):
        if self.errors:
            return self.cleaned_data

        import_file = self.cleaned_data["import_file"]

        try:
            data = _get_parsed_data(import_file)
        except (ValueError, TypeError):
            raise ValidationError("File is not valid")

        first_item = data[0]
        is_placeholder = isinstance(first_item, ArchivedPlaceholder)
        page_import = bool(self.cleaned_data["cms_pagecontent"])
        plugins_import = not page_import

        if (is_placeholder and plugins_import) or (page_import and not is_placeholder):
            raise ValidationError("Incorrect json format used.")

        self.cleaned_data["import_data"] = data
        return self.cleaned_data

    def run_import(self):
        data = self.cleaned_data
        language = data["language"]
        target_page = data["cms_pagecontent"]
        target_plugin = data["plugin"]
        target_placeholder = data["placeholder"]

        if target_plugin:
            target_plugin_id = target_plugin.pk
            target_placeholder = target_plugin.placeholder
        else:
            target_plugin_id = None

        if target_page:
            import_plugins_to_page(
                placeholders=data["import_data"],
                page=target_page,
                language=language,
            )
        else:
            import_plugins(
                plugins=data["import_data"],
                placeholder=target_placeholder,
                language=language,
                root_plugin_id=target_plugin_id,
            )



__version__ = "2.0.0a"

default_app_config = "djangocms_transfer.apps.TranferConfig"


def get_serializer_name(default="python"):
    from django.conf import settings

    return getattr(settings, "DJANGO_CMS_TRANSFER_SERIALIZER", default)



from collections import defaultdict

from django.conf import settings
from django.core import serializers

from . import get_serializer_name
from .utils import get_plugin_fields, get_plugin_model


def get_bound_plugins(plugins):
    # COMPAT: CMS<3.5
    # When we drop support for CMS3.4 we can replace all this method by
    # from cms.utils.plugins import get_bound_plugins
    plugin_types_map = defaultdict(list)
    plugin_ids = []
    plugin_lookup = {}

    # make a map of plugin types, needed later for downcasting
    for plugin in plugins:
        plugin_ids.append(plugin.pk)
        plugin_types_map[plugin.plugin_type].append(plugin.pk)

    for plugin_type, pks in plugin_types_map.items():
        plugin_model = get_plugin_model(plugin_type)
        plugin_queryset = plugin_model.objects.filter(pk__in=pks)

        # put them in a map so we can replace the base CMSPlugins with their
        # downcasted versions
        for instance in plugin_queryset.iterator():
            plugin_lookup[instance.pk] = instance

    for plugin in plugins:
        parent_not_available = (
            not plugin.parent_id or plugin.parent_id not in plugin_ids
        )
        # The plugin either has no parent or needs to have a non-ghost parent
        valid_parent = parent_not_available or plugin.parent_id in plugin_lookup

        if valid_parent and plugin.pk in plugin_lookup:
            yield plugin_lookup[plugin.pk]


def get_plugin_data(plugin, only_meta=False):
    if only_meta:
        custom_data = None
    else:
        plugin_fields = get_plugin_fields(plugin.plugin_type)
        _plugin_data = serializers.serialize(
            get_serializer_name(), (plugin,), fields=plugin_fields
        )[0]
        custom_data = _plugin_data["fields"]

    plugin_data = {
        "pk": plugin.pk,
        "creation_date": plugin.creation_date,
        "position": plugin.position,
        "plugin_type": plugin.plugin_type,
        "parent_id": plugin.parent_id,
        "data": custom_data,
    }

    gpd = getattr(settings, "DJANGOCMS_TRANSFER_PROCESS_EXPORT_PLUGIN_DATA", None)
    if gpd:
        module, function = gpd.rsplit(".", 1)
        return getattr(__import__(module, fromlist=[""]), function)(plugin, plugin_data)
    else:
        return plugin_data



from collections import namedtuple

from django.core.serializers import deserialize
from django.db import transaction
from django.utils.encoding import force_str
from django.utils.functional import cached_property

from cms.api import add_plugin
from cms.models import CMSPlugin

from . import get_serializer_name
from .utils import get_plugin_model


BaseArchivedPlugin = namedtuple(
    "ArchivedPlugin",
    ["pk", "creation_date", "position", "plugin_type", "parent_id", "data"],
)

ArchivedPlaceholder = namedtuple("ArchivedPlaceholder", ["slot", "plugins"])


class ArchivedPlugin(BaseArchivedPlugin):
    @cached_property
    def model(self):
        return get_plugin_model(self.plugin_type)

    @cached_property
    def deserialized_instance(self):
        data = {
            "model": force_str(self.model._meta),
            "fields": self.data,
        }

        # TODO: Handle deserialization error
        return list(deserialize(get_serializer_name(), [data]))[0]

    @transaction.atomic
    def restore(self, placeholder, language, parent=None):
        m2m_data = {}
        data = self.data.copy()

        if self.model is not CMSPlugin:
            fields = self.model._meta.get_fields()
            for field in fields:
                if field.related_model is not None:
                    if field.many_to_many:
                        if data.get(field.name):
                            m2m_data[field.name] = data[field.name]
                        data.pop(field.name, None)
                    elif data.get(field.name):
                        try:
                            obj = field.related_model.objects.get(pk=data[field.name])
                        except field.related_model.DoesNotExist:
                            obj = None
                        data[field.name] = obj

        plugin = add_plugin(
            placeholder,
            self.plugin_type,
            language,
            position="last-child",
            target=parent,
            **data,
        )

        if self.model is not CMSPlugin:
            fields = self.model._meta.get_fields()
            for field in fields:
                if field.related_model is not None and m2m_data.get(field.name):
                    if field.many_to_many:
                        objs = field.related_model.objects.filter(
                            pk__in=m2m_data[field.name]
                        )
                        attr = getattr(plugin, field.name)
                        attr.set(objs)

        return plugin
